
#pragma once

// Genode includes
#include <base/log.h>
#include <libc/component.h>
#include <timer_session/connection.h>

// libc includes
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

// stdcxx includes
#include <functional>

// gapfruit includes
#include <util/hexdump.h>

// local includes
#include "exception.h"


namespace Hsm {

	using Genode::Env;
	using Genode::error;
	using Genode::log;
	using Genode::addr_t;

	using Ip_address  = Genode::String<64>;

	using Callback    = std::function<void()>;

	class Socket_wrapper;
};


class Hsm::Socket_wrapper
{
	private:
		using Signal_handler = Genode::Signal_handler<Socket_wrapper>;

		enum {
			INVALID_SOCKET = -1,
			RETRY_TIME = 2000 * 1000,     // time between reconnect tries in microseconds
		};

		Env&                       _env;
		Signal_handler             _timer_handler    { _env.ep(),
			                                           *this,
			                                           &Socket_wrapper::_handle_timer };
		mutable Timer::Connection  _timer            { _env };
		mutable int                _socket           { INVALID_SOCKET };
		mutable bool               _connected        { false };

		const Ip_address           _ip_address;
		const uint16_t             _port;
		const Callback             _connected_cb;
		const Callback             _disconnected_cb;

		void _handle_timer()
		{
			if (_connected) return;

			Libc::with_libc([&] () {

				log("connecting to HSM at ", _ip_address, ":", _port);

				_socket = socket(AF_INET , SOCK_STREAM , 0);
				if (_socket == INVALID_SOCKET) {
					error("socket failed will retry later.");
					_timer.trigger_once(RETRY_TIME);
					return;
				}

				struct sockaddr_in address;
				address.sin_addr.s_addr = inet_addr(_ip_address.string());
				address.sin_family      = AF_INET;
				address.sin_port        = htons(_port);

				if (connect(_socket, (struct sockaddr *)&address, sizeof(address)) < 0) {
					close(_socket);
					_socket = INVALID_SOCKET;
					error("connect failed will retry later.");
					_timer.trigger_once(RETRY_TIME);
					return;
				}

				_connected = true;
				log("socket connected");

				if(_connected_cb != nullptr) {
					_connected_cb();
				}
			});
		}

		void _close_on_error() const
		{
			error("socket error during communication");

			close(_socket);
			_socket = INVALID_SOCKET;

			_connected = false;
			if (_disconnected_cb != nullptr) {
				_disconnected_cb();
			}

			// try to reconnect
			_timer.trigger_once(2);
		}

	public:

		Socket_wrapper(Env& env,
		               const Ip_address& ip_address,
		               const uint16_t port,
		               Callback connected_cb,
		               Callback disconnected_cb) :
			_env             { env },
			_ip_address      { ip_address },
			_port            { port },
			_connected_cb    { connected_cb },
			_disconnected_cb { disconnected_cb }
		{
			_timer.sigh(_timer_handler);

			// try to connect to HSM
			_handle_timer();
		}

		~Socket_wrapper() = default;

		size_t write(const void* data, size_t size) const
		{
			if (0 == size) return size;

			if (!_connected) {
				error("not connected to HSM");
				throw Remote_hsm_error { };
			}

			Libc::with_libc([&] () {
				size_t buf_fill = 0;
				while (buf_fill < size) {
					ssize_t part = ::write(_socket, (void *)((addr_t)data + buf_fill), size - buf_fill);
					if (part < 0) {
						error("failed to write to HSM. errno=", errno);
						_close_on_error();
						throw Remote_hsm_error { };
					}
					if (part == 0) {
						error("failed to write to HSM. EOF");
						_close_on_error();
						throw Remote_hsm_error { };
					}
					buf_fill += part;
				}
			});

			return size;
		}

		size_t read(void* data, size_t size) const
		{
			if (!_connected) {
				error("not connected to HSM");
				throw Remote_hsm_error { };
			}

			Libc::with_libc([&] () {
				size_t buf_fill = 0;
				while (buf_fill < size) {
					ssize_t part = ::read(_socket, (void *)((addr_t)data + buf_fill), size - buf_fill);
					if (part < 0) {
						error("failed to read from HSM. errno=", errno);
						_close_on_error();
						throw Remote_hsm_error { };
					}
					if (part == 0) {
						error("failed to read from HSM. EOF");
						_close_on_error();
						throw Remote_hsm_error { };
					}
					buf_fill += part;
				}
			});

			return size;
		}
};
